package it

import (
	"context"
	"encoding/json"
	"testing"
	"time"

	"github.com/coder/websocket"
	"github.com/google/go-cmp/cmp"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v18/internal/module/kubernetes_api/rpc"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v18/internal/tool/testing/testhelpers"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v18/pkg/agentcfg"
	"google.golang.org/protobuf/encoding/protojson"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/utils/ptr"
)

func TestGraphAPIManual(t *testing.T) {
	t.SkipNow()

	const (
		kasKubernetesAPIURL = "http://gdk.test:8154/-/k8s-proxy"
		credential          = "Bearer pat:3:tL1uN9hQTcZgJ2yUpTKM" // Oh no a real GDK PAT!
	)

	ctx, cancel := context.WithTimeout(context.Background(), time.Minute)
	defer cancel()
	conn, err := dialWS(ctx, kasKubernetesAPIURL+"/graph", credential, "gitlab-agent-graph-api")
	require.NoError(t, err)

	gw := graphWatcher{
		t:       t,
		conn:    conn,
		require: require.New(t),
		assert:  assert.New(t),
	}
	defer gw.Close()

	reqData, err := protojson.Marshal(&rpc.WatchGraphWebSocketRequest{
		Queries: []*rpc.Query{
			{
				Query: &rpc.Query_Include_{
					Include: &rpc.Query_Include{
						ResourceSelectorExpression: "group == '' && version == 'v1' && resource == 'pods'",
						Object: &rpc.Query_IncludeObject{
							ObjectSelectorExpression: "'k8s-app' in labels && labels['k8s-app'] == 'kube-dns'",
							JsonPath:                 "['metadata.name', 'metadata.namespace']",
						},
					},
				},
			},
		},
		Namespaces: &rpc.Namespaces{
			Names: []string{"kube-system"},
		},
	})
	require.NoError(t, err)

	err = conn.Write(ctx, websocket.MessageText, reqData)
	require.NoError(t, err)

	for {
		//ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
		a := gw.nextAction(ctx)
		aBytes, err := json.Marshal(a)
		require.NoError(t, err)
		t.Log(string(aBytes))
		//cancel()
	}
}

func (s *integrationSuite) TestGraphAPI() {
	gitalySrv := s.startGitaly(gitalyOptions{})
	s.setupGitLabMocks(gitalySrv, true)
	s.setupGitalyMocksForAgentConfig(&agentcfg.ConfigurationFile{
		Observability: &agentcfg.ObservabilityCF{
			Logging: &agentcfg.LoggingCF{
				Level:     agentcfg.LogLevelEnum_debug,
				GrpcLevel: ptr.To(agentcfg.LogLevelEnum_debug),
			},
		},
	})

	gitLabSrv := s.startGitLab(gitLabOptions{})

	kas := s.startKAS(kasOptions{
		name:       "kas",
		cfg:        s.baseKASConfig(gitLabSrv),
		gitalyPort: gitalySrv.port,
	})

	s.startAgentk(agentkOptions{
		kasURL: kas.agentServerURL,
		token:  agentkToken1,
		name:   "agent1",
		ns:     s.ns,
	})

	s.Run("create,delete,owner references,whole object,namespace by label", func() {
		gw := s.newGraphWatcher(kas, &rpc.WatchGraphWebSocketRequest{
			Queries: []*rpc.Query{
				{
					Query: &rpc.Query_Include_{
						Include: &rpc.Query_Include{
							ResourceSelectorExpression: "group == '' && version == 'v1' && resource == 'configmaps'",
							Object: &rpc.Query_IncludeObject{
								LabelSelector: "ls=test-cm1",
							},
						},
					},
				},
			},
			Namespaces: &rpc.Namespaces{
				LabelSelector: "ls=" + nsLabel,
			},
		})
		defer gw.Close()

		_, err := s.k8s.CoreV1().ConfigMaps(s.ns).Create(s.ctx, &corev1.ConfigMap{
			TypeMeta: metav1.TypeMeta{
				Kind:       "ConfigMap",
				APIVersion: "v1",
			},
			ObjectMeta: metav1.ObjectMeta{
				Name:      "cm1",
				Namespace: s.ns,
				Labels: map[string]string{
					"ls": "test-cm1",
				},
				OwnerReferences: []metav1.OwnerReference{
					{
						APIVersion: "v1",
						Kind:       "Namespace",
						Name:       s.ns,
						UID:        s.nsObj.UID,
						Controller: ptr.To(true),
					},
				},
			},
			Data: map[string]string{
				"key": "value",
			},
		}, metav1.CreateOptions{})
		s.Require().NoError(err)

		a := gw.nextAction(s.ctx)
		cmVx := vertex{
			Group:     "",
			Version:   "v1",
			Resource:  "configmaps",
			Namespace: s.ns,
			Name:      "cm1",
		}
		s.assertEqualSetVertex(a, &jsonSetVertex{
			Vertex: cmVx,
			Object: map[string]any{
				"apiVersion": "v1",
				"kind":       "ConfigMap",
				"metadata": map[string]any{
					"name":      "cm1",
					"namespace": s.ns,
					"labels": map[string]any{
						"ls": "test-cm1",
					},
					"ownerReferences": []any{
						map[string]any{
							"apiVersion": "v1",
							"kind":       "Namespace",
							"name":       s.ns,
							"uid":        string(s.nsObj.UID),
							"controller": true,
						},
					},
				},
				"data": map[string]any{
					"key": "value",
				},
			},
		})

		owArc := Arc{
			Source: cmVx,
			Destination: vertex{
				Group:    "",
				Version:  "v1",
				Resource: "namespaces",
				Name:     s.ns,
			},
			Type: "or",
		}
		a = gw.nextAction(s.ctx)
		s.assertEqualSetArc(a, &jsonSetArc{
			Arc: owArc,
			Attributes: map[string]any{
				"c": true,
			},
		})

		err = s.k8s.CoreV1().ConfigMaps(s.ns).Delete(s.ctx, "cm1", metav1.DeleteOptions{})
		s.Require().NoError(err)

		a = gw.nextAction(s.ctx)
		s.assertEqualDeleteArc(a, &jsonDeleteArc{
			Arc: owArc,
		})

		a = gw.nextAction(s.ctx)
		s.assertEqualDeleteVertex(a, &jsonDeleteVertex{
			Vertex: cmVx,
		})
	})
	s.Run("create,JSON path,namespace by name", func() {
		gw := s.newGraphWatcher(kas, &rpc.WatchGraphWebSocketRequest{
			Queries: []*rpc.Query{
				{
					Query: &rpc.Query_Include_{
						Include: &rpc.Query_Include{
							ResourceSelectorExpression: "group == '' && version == 'v1' && resource == 'configmaps'",
							Object: &rpc.Query_IncludeObject{
								LabelSelector: "ls=test-cm2",
								JsonPath:      "['metadata.name', 'metadata.namespace']",
							},
						},
					},
				},
			},
			Namespaces: &rpc.Namespaces{
				Names: []string{s.ns},
			},
		})
		defer gw.Close()

		_, err := s.k8s.CoreV1().ConfigMaps(s.ns).Create(s.ctx, &corev1.ConfigMap{
			TypeMeta: metav1.TypeMeta{
				Kind:       "ConfigMap",
				APIVersion: "v1",
			},
			ObjectMeta: metav1.ObjectMeta{
				Name:      "cm2",
				Namespace: s.ns,
				Labels: map[string]string{
					"ls": "test-cm2",
				},
			},
		}, metav1.CreateOptions{})
		s.Require().NoError(err)

		a := gw.nextAction(s.ctx)
		s.assertEqualSetVertex(a, &jsonSetVertex{
			Vertex: vertex{
				Group:     "",
				Version:   "v1",
				Resource:  "configmaps",
				Namespace: s.ns,
				Name:      "cm2",
			},
			JSONPath: []any{
				"cm2", s.ns,
			},
		})
	})
}

func (s *integrationSuite) newGraphWatcher(kas *kasHolder, req *rpc.WatchGraphWebSocketRequest) *graphWatcher {
	reqData, err := protojson.Marshal(req)
	s.Require().NoError(err)

	conn := s.dialWS(kas, testhelpers.AgentID)

	err = conn.Write(s.ctx, websocket.MessageText, reqData)
	if err != nil {
		_ = conn.CloseNow()
	}
	s.Require().NoError(err)

	return &graphWatcher{
		t:       s.T(),
		conn:    conn,
		require: s.Require(),
		assert:  s.Assert(),
	}
}

func (s *integrationSuite) assertEqualSetVertex(a jsonWatchGraphAction, sv *jsonSetVertex) {
	if a.SetVertex != nil && a.SetVertex.Object != nil {
		md := a.SetVertex.Object["metadata"].(map[string]any)
		delete(md, "creationTimestamp")
		delete(md, "managedFields")
		delete(md, "resourceVersion")
		delete(md, "uid")
	}
	s.assertEmptyDiff(a, jsonWatchGraphAction{
		SetVertex: sv,
	})
}

func (s *integrationSuite) assertEqualDeleteVertex(a jsonWatchGraphAction, sv *jsonDeleteVertex) {
	s.assertEmptyDiff(a, jsonWatchGraphAction{
		DeleteVertex: sv,
	})
}

func (s *integrationSuite) assertEqualSetArc(a jsonWatchGraphAction, sa *jsonSetArc) {
	s.assertEmptyDiff(a, jsonWatchGraphAction{
		SetArc: sa,
	})
}

func (s *integrationSuite) assertEqualDeleteArc(a jsonWatchGraphAction, da *jsonDeleteArc) {
	s.assertEmptyDiff(a, jsonWatchGraphAction{
		DeleteArc: da,
	})
}

func (s *integrationSuite) assertEmptyDiff(a, b any) {
	s.T().Helper()
	s.Empty(cmp.Diff(a, b))
}

type graphWatcher struct {
	t       *testing.T
	conn    *websocket.Conn
	require *require.Assertions
	assert  *assert.Assertions
	resp    jsonWatchGraphResponse
}

func (w *graphWatcher) Close() {
	err := w.conn.Close(websocket.StatusNormalClosure, "")
	w.assert.NoError(err)
}

func (w *graphWatcher) nextAction(ctx context.Context) jsonWatchGraphAction {
	for {
		if len(w.resp.Actions) > 0 {
			a := w.resp.Actions[0]
			w.resp.Actions = w.resp.Actions[1:]
			return a
		}
		_, msg, err := w.conn.Read(ctx)
		w.require.NoError(err)

		err = json.Unmarshal(msg, &w.resp)
		w.require.NoError(err)
		for _, warn := range w.resp.Warnings {
			warnData, err := json.Marshal(warn)
			w.require.NoError(err)
			w.t.Log(string(warnData))
		}
	}
}

type jsonWatchGraphResponse struct {
	Actions  []jsonWatchGraphAction  `json:"actions,omitempty"`
	Warnings []jsonWatchGraphWarning `json:"warnings,omitempty"`
}

type jsonWatchGraphAction struct {
	SetVertex    *jsonSetVertex    `json:"svx,omitempty"`
	DeleteVertex *jsonDeleteVertex `json:"dvx,omitempty"`
	SetArc       *jsonSetArc       `json:"sarc,omitempty"`
	DeleteArc    *jsonDeleteArc    `json:"darc,omitempty"`
}

type vertex struct {
	Group     string `json:"g,omitempty"`
	Version   string `json:"v"`
	Resource  string `json:"r"`
	Namespace string `json:"ns,omitempty"`
	Name      string `json:"n"`
}

type jsonSetVertex struct {
	Vertex   vertex         `json:"vx"`
	Object   map[string]any `json:"o,omitempty"`
	JSONPath []any          `json:"j,omitempty"`
}

type jsonDeleteVertex struct {
	Vertex vertex `json:"vx"`
}

type Arc struct {
	Source      vertex `json:"s"`
	Destination vertex `json:"d"`
	Type        string `json:"t"`
}

type jsonSetArc struct {
	Arc
	Attributes map[string]any `json:"a,omitempty"`
}

type jsonDeleteArc struct {
	Arc
}

type jsonWatchGraphWarning struct {
	Type       string         `json:"t"`
	Message    string         `json:"m"`
	Attributes map[string]any `json:"a,omitempty"`
}
