package watch_graph //nolint:stylecheck

import (
	"context"
	"fmt"
	"log/slog"
	"regexp"
	"strings"
	"time"

	"github.com/fvbommel/sortorder"
	"github.com/google/cel-go/common/types"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v18/internal/tool/logz"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime/schema"
	"k8s.io/client-go/discovery"
)

const (
	discoveryPollPeriod = time.Minute
)

var (
	isStableRegex = regexp.MustCompile(`^v\d+$`)
)

type gvrInfoFiltered struct {
	gvrInfo
	kind string
}

type wgDiscovery struct {
	log         *slog.Logger
	queries     []any // Elements are QueryInclude or QueryExclude
	discoClient discovery.AggregatedDiscoveryInterface
	gvrCh       chan<- gvrInfos
	errCh       chan<- Error   // fatal errors
	warnCh      chan<- Warning // warnings
}

func (d *wgDiscovery) Run(ctx context.Context) {
	d.doDiscovery(ctx)
	t := time.NewTicker(discoveryPollPeriod)
	defer t.Stop()
	done := ctx.Done()
	for {
		select {
		case <-done:
			return
		case <-t.C:
			d.doDiscovery(ctx)
		}
	}
}

func (d *wgDiscovery) doDiscovery(ctx context.Context) {
	partial := false
	done := ctx.Done()
	_, resources, err := discovery.ServerGroupsAndResources(d.discoClient)
	switch err.(type) { //nolint:errorlint
	case nil: // no error
	case *discovery.ErrGroupDiscoveryFailed: // partial data
		select {
		case <-done:
			return
		case d.warnCh <- Warning{
			Type:    discoFailedWarnType,
			Message: fmt.Sprintf("discovery partially failed: %v", err),
		}:
		}
		partial = true
	default: // something is wrong
		select {
		case <-done:
			return
		case d.warnCh <- Warning{
			Type:    discoFailedWarnType,
			Message: fmt.Sprintf("discovery failed: %v", err),
		}:
		}
		return
	}

	infos, gk2info, err := d.filterDiscoveredResources(ctx, resources)
	if err != nil {
		select {
		case <-done:
		case d.errCh <- Error{
			Message: err.Error(),
			Code:    InvalidArgument,
		}:
		}
		return
	}
	select {
	case <-done:
	case d.gvrCh <- gvrInfos{infos: infos, gk2info: gk2info, isPartial: partial}:
	}
}

// All returned errors are fatal errors.
func (d *wgDiscovery) filterDiscoveredResources(ctx context.Context, resources []*metav1.APIResourceList) (map[schema.GroupVersionResource]gvrInfo, map[schema.GroupKind]gvrInfoLite, error) {
	gr2infos := map[schema.GroupResource][]gvrInfoFiltered{} // group+resource -> []infos for the resources of various versions in this group
	gk2info := map[schema.GroupKind]gvrInfoLite{}
	for _, resList := range resources {
		err := d.filterDiscoveredGV(ctx, resList, gr2infos, gk2info)
		if err != nil {
			return nil, nil, err
		}
	}

	infos := make(map[schema.GroupVersionResource]gvrInfo, len(gr2infos))
	for _, inf := range gr2infos {
		selectedInf := selectVersion(inf)
		infos[selectedInf.gvr] = selectedInf.gvrInfo
		gk := schema.GroupKind{
			Group: selectedInf.gvr.Group,
			Kind:  selectedInf.kind,
		}
		gk2info[gk] = gvrInfoLite{
			gvr:        selectedInf.gvr,
			namespaced: selectedInf.namespaced,
		}
	}

	return infos, gk2info, nil
}

func (d *wgDiscovery) filterDiscoveredGV(ctx context.Context, resList *metav1.APIResourceList,
	gr2infos map[schema.GroupResource][]gvrInfoFiltered, gk2info map[schema.GroupKind]gvrInfoLite) error {
	gv, err := schema.ParseGroupVersion(resList.GroupVersion)
	if err != nil {
		return err
	}
	d.log.Debug("Filtering discovered resources for group/version", logz.GV(gv))
nextResource:
	for _, res := range resList.APIResources {
		if isSubresource(res.Name) {
			d.log.Debug("Skipping subresource", logz.GVR(gv.WithResource(res.Name)))
			continue
		}
		gk := schema.GroupKind{
			Group: gv.Group,
			Kind:  res.Kind,
		}
		gk2info[gk] = gvrInfoLite{
			gvr: schema.GroupVersionResource{
				Group:    gv.Group,
				Version:  "", // to be selected later
				Resource: res.Name,
			},
			namespaced: res.Namespaced,
		}
		vars := map[string]any{
			"group":      gv.Group,
			"version":    gv.Version,
			"resource":   res.Name,
			"namespaced": res.Namespaced,
		}
		for pos, query := range d.queries {
			switch q := query.(type) {
			case QueryInclude:
				val, _, err := q.ResourceSelectorExpression.ContextEval(ctx, vars)
				if err != nil {
					return fmt.Errorf("queries[%d] include expression: %w", pos, err)
				}
				if val.(types.Bool) { // match! this GVR is included.
					gvr := gv.WithResource(res.Name)
					gr := gvr.GroupResource()
					gr2infos[gr] = append(gr2infos[gr], gvrInfoFiltered{
						gvrInfo: gvrInfo{
							gvr:        gvr,
							object:     q.Object,
							namespaced: res.Namespaced,
						},
						kind: res.Kind,
					})
					continue nextResource
				}
			case QueryExclude:
				val, _, err := q.ResourceSelectorExpression.ContextEval(ctx, vars)
				if err != nil {
					return fmt.Errorf("queries[%d] exclude expression: %w", pos, err)
				}
				if val.(types.Bool) { // match! this GVR is excluded.
					continue nextResource
				}
			default:
				return fmt.Errorf("unknown query type: %T", q) // should never happen
			}
			// no match, try next query
		}
		// no queries matched, ignoring this resource
	}
	return nil
}

func selectVersion(inf []gvrInfoFiltered) gvrInfoFiltered {
	selected := inf[0] // we know it's not empty
	if len(inf) == 1 {
		return selected
	}
	selectedIsStable := isStableVersion(selected.gvr.Version)

	for _, maybeNewSelected := range inf[1:] {
		maybeNewSelectedIsStable := isStableVersion(maybeNewSelected.gvr.Version)
		switch {
		case !selectedIsStable && maybeNewSelectedIsStable: // unstable -> stable upgrade
			selected = maybeNewSelected
			selectedIsStable = true
			continue
		case selectedIsStable && !maybeNewSelectedIsStable: // ignore unstable when stable is selected
			continue
		default: // compare stable vs stable or unstable vs unstable
			if sortorder.NaturalLess(selected.gvr.Version, maybeNewSelected.gvr.Version) {
				selected = maybeNewSelected
				selectedIsStable = maybeNewSelectedIsStable
			}
		}
	}
	return selected
}

func isSubresource(name string) bool {
	return strings.IndexByte(name, '/') != -1
}

func isStableVersion(ver string) bool {
	return isStableRegex.MatchString(ver)
}
