package watch_graph //nolint:stylecheck

import (
	"context"
	"fmt"

	"github.com/google/cel-go/common/types"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/runtime/schema"
	"k8s.io/client-go/tools/cache"
)

type resourceEventHandler struct {
	ctx          context.Context
	info         gvrInfo
	setVertex    func(VertexID, VertexData, arcSet[VertexID, ArcType, ArcData])
	deleteVertex func(VertexID)
	gk2info      func(schema.GroupKind) gvrInfoLite // returns an empty struct if mapping was not found
	emitError    func(Error)
	emitWarning  func(Warning)
}

func (h *resourceEventHandler) OnAdd(obj any, isInInitialList bool) {
	h.maybeSetVertex(obj)
}

func (h *resourceEventHandler) OnUpdate(oldObj, newObj any) {
	h.maybeSetVertex(newObj)
}

func (h *resourceEventHandler) OnDelete(obj any) {
	if x, ok := obj.(cache.DeletedFinalStateUnknown); ok {
		obj = x.Obj
	}
	o := obj.(*unstructured.Unstructured)
	h.deleteVertex(VertexID{
		GVR:       h.info.gvr,
		Namespace: o.GetNamespace(),
		Name:      o.GetName(),
	})
}

func (h *resourceEventHandler) maybeSetVertex(obj any) {
	add := true
	o := obj.(*unstructured.Unstructured)
	if h.info.object.ObjectSelectorExpression != nil {
		val, _, err := h.info.object.ObjectSelectorExpression.ContextEval(h.ctx, objectToEvalVars(o, h.info.gvr))
		if err != nil {
			h.emitError(Error{
				Message: fmt.Sprintf("query.include: object selector expression failed for %s %s/%s: %v", h.info.gvr, o.GetNamespace(), o.GetName(), err),
				Code:    InvalidArgument,
			})
			return
		}
		add = bool(val.(types.Bool))
	}
	vid := VertexID{
		GVR:       h.info.gvr,
		Namespace: o.GetNamespace(),
		Name:      o.GetName(),
	}
	if add {
		vx := wgObject2vertex{
			info:    h.info,
			gk2info: h.gk2info,
		}
		vd, arcs, gWarns, gErr := vx.Get(o)
		for _, gWarn := range gWarns {
			h.emitWarning(gWarn)
		}
		if gErr != nil {
			h.emitError(*gErr)
			return
		}
		h.setVertex(vid, vd, arcs)
	} else {
		h.deleteVertex(vid)
	}
}
