package watch_graph //nolint:stylecheck

import (
	"context"
	"errors"
	"log/slog"
	"sync"
	"sync/atomic"

	"github.com/google/cel-go/cel"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v18/internal/module/kubernetes_api/rpc"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v18/internal/module/modshared"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v18/internal/tool/fieldz"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v18/internal/tool/logz"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/runtime/schema"
	"k8s.io/apimachinery/pkg/util/sets"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/apimachinery/pkg/watch"
	"k8s.io/client-go/discovery"
	"k8s.io/client-go/dynamic"
	"k8s.io/client-go/tools/cache"
	"k8s.io/client-go/util/jsonpath"
)

type ArcType byte

const (
	ownerReferenceArcType ArcType = iota
	labelSelectorArcType
	referenceArcType
	transitiveReferenceArcType
	viaReferenceArcType
)

func (at ArcType) String() string {
	switch at {
	case ownerReferenceArcType:
		return "or"
	case labelSelectorArcType:
		return "ls"
	case referenceArcType:
		return "r"
	case transitiveReferenceArcType:
		return "tr"
	case viaReferenceArcType:
		return "via"
	default:
		return "unknown"
	}
}

type gvrInfoLite struct {
	gvr        schema.GroupVersionResource
	namespaced bool
}

type gvrInfo struct {
	gvr        schema.GroupVersionResource
	object     QueryIncludeObject
	namespaced bool
}

type gvrInfos struct {
	infos map[schema.GroupVersionResource]gvrInfo
	// Our little simple replacement for the rest mapper.
	// It contains all mappings that we got from discovery, not only those selected by the queries.
	// Version selected by queries is preferred to other versions of the same group+resource.
	// If no version was selected, the server-preferred version is used.
	gk2info map[schema.GroupKind]gvrInfoLite

	// Is discovery info partial or complete.
	// Partial info may not contain all GVRs because e.g. one of the configured aggregated API servers is offline
	// and hence discovery info for the types it exposes is unavailable at the moment.
	// See https://kubernetes.io/docs/concepts/extend-kubernetes/api-extension/apiserver-aggregation/.
	isPartial bool
}

// QueryIncludeObject holds config for object filtering.
type QueryIncludeObject struct {
	LabelSelector            string
	FieldSelector            string
	ObjectSelectorExpression cel.Program        // can be nil
	JSONPath                 *jsonpath.JSONPath // can be nil
}

type QueryInclude struct {
	ResourceSelectorExpression cel.Program // never nil

	Object QueryIncludeObject
}

type QueryExclude struct {
	ResourceSelectorExpression cel.Program // never nil
}

type Namespaces struct {
	Names                    []string
	LabelSelector            string
	FieldSelector            string
	ObjectSelectorExpression cel.Program // can be nil
}

// All checks if all namespaces should be watched.
func (n *Namespaces) All() bool {
	return len(n.Names) == 0 && n.LabelSelector == "" && n.ObjectSelectorExpression == nil
}

type selectedGVR struct {
	//inf  cache.SharedIndexInformer
	stop context.CancelFunc
	wait func()
}

type VertexID struct {
	GVR       schema.GroupVersionResource
	Namespace string
	Name      string
}

type VertexData struct {
	Object   []byte
	JSONPath []byte
}

type ArcData []byte

type Opts struct {
	Log         *slog.Logger
	Queries     []any // Elements are QueryInclude or QueryExclude
	Namespaces  Namespaces
	Roots       *rpc.Roots // can be nil
	RPCAPI      modshared.RPCAPI
	DiscoClient discovery.AggregatedDiscoveryInterface
	Client      dynamic.Interface
	OnWarning   func(Warning)
	Graph       *ObjectGraph[VertexID, ArcType, VertexData, ArcData]
}

type WatchGraph struct {
	opts    Opts
	graphMu sync.Mutex
	errCh   chan Error   // fatal errors
	warnCh  chan Warning // warnings
	allNS   bool

	// Fields below are mutable but are only accessed from the Run goroutine.

	wg           wait.Group
	selectedGVRs map[schema.GroupVersionResource]map[string]selectedGVR // GVR -> namespace -> data
	selectedNS   sets.Set[string]
	gvrInfos     atomic.Pointer[gvrInfos]
}

func New(opts Opts) *WatchGraph {
	g := &WatchGraph{
		opts:         opts,
		errCh:        make(chan Error),
		warnCh:       make(chan Warning),
		allNS:        opts.Namespaces.All(),
		selectedGVRs: map[schema.GroupVersionResource]map[string]selectedGVR{},
		selectedNS:   sets.Set[string]{},
	}
	g.gvrInfos.Store(&gvrInfos{})
	return g
}

func (g *WatchGraph) Run(ctx context.Context) *Error {
	defer g.wg.Wait()
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	var nsAddCh chan string
	var nsDelCh chan string
	gvrCh := make(chan gvrInfos)

	if !g.allNS {
		nsAddCh = make(chan string)
		nsDelCh = make(chan string)
		g.wg.Start(func() {
			wgn := wgNamespaces{
				namespaces: g.opts.Namespaces,
				client:     g.opts.Client,
				nsAddCh:    nsAddCh,
				nsDelCh:    nsDelCh,
				errCh:      g.errCh,
				warnCh:     g.warnCh,
			}
			wgn.Run(ctx)
		})
	}

	g.wg.Start(func() {
		d := wgDiscovery{
			log:         g.opts.Log,
			queries:     g.opts.Queries,
			discoClient: g.opts.DiscoClient,
			gvrCh:       gvrCh,
			errCh:       g.errCh,
			warnCh:      g.warnCh,
		}
		d.Run(ctx)
	})

	done := ctx.Done()
	for {
		select {
		case <-done:
			return nil
		case ns := <-nsAddCh:
			g.handleNamespaceAdd(ctx, ns)
		case ns := <-nsDelCh:
			g.handleNamespaceDel(ns)
		case gvr := <-gvrCh:
			g.handleGVRUpdate(ctx, gvr)
		case err := <-g.errCh:
			return &err
		case warn := <-g.warnCh:
			g.opts.RPCAPI.HandleProcessingError(g.opts.Log, "Graph API warning", errors.New(warn.Message),
				fieldz.Attributes(warn.Attributes))
			g.opts.OnWarning(warn)
		}
	}
}

func (g *WatchGraph) handleGVRUpdate(ctx context.Context, infos gvrInfos) {
	// 1. Assign the field before starting any informers since the callbacks can access it.
	g.gvrInfos.Store(&infos)

	// 2. Start/stop informers as necessary.
	newSelectedGVRs := make(map[schema.GroupVersionResource]map[string]selectedGVR, len(infos.infos))
	for gvr, info := range infos.infos {
		existingData, ok := g.selectedGVRs[gvr]
		if ok { // no changes, keep everything as-is, just move
			newSelectedGVRs[gvr] = existingData
			delete(g.selectedGVRs, gvr) // remove from current set
		} else { // new GVR! Need to start informers for all/selected namespaces.
			if !info.namespaced || g.allNS {
				newSelectedGVRs[gvr] = map[string]selectedGVR{
					metav1.NamespaceAll: g.startInformerForNamespacedGVR(ctx, info, metav1.NamespaceAll),
				}
			} else {
				newSelectedGVRs[gvr] = make(map[string]selectedGVR, g.selectedNS.Len())
				for ns := range g.selectedNS {
					newSelectedGVRs[gvr][ns] = g.startInformerForNamespacedGVR(ctx, info, ns)
				}
			}
		}
	}
	var waits []func()
	for _, ns2data := range g.selectedGVRs {
		for _, data := range ns2data {
			data.stop()
			waits = append(waits, data.wait)
		}
	}
	for _, w := range waits {
		w()
	}
	g.selectedGVRs = newSelectedGVRs
}

func (g *WatchGraph) startInformerForNamespacedGVR(ctx context.Context, info gvrInfo, ns string) selectedGVR {
	var waitWg sync.WaitGroup
	infCtx, infCancel := context.WithCancel(ctx)
	done := infCtx.Done()
	nsInf := newInformer(g.opts.Client, info.gvr, ns, info.object.LabelSelector, info.object.FieldSelector) //nolint:contextcheck
	_, err := nsInf.AddEventHandler(g.resourceEventHandlerForGVR(ctx, info))
	if err != nil { // cannot happen, but let's handle it gracefully
		g.opts.Log.Error("Failed to add event handler", logz.Error(err))
		return selectedGVR{
			stop: infCancel,
			wait: waitWg.Wait,
		}
	}
	//nsInf.HasSynced() // TODO handle stuck sync
	waitWg.Add(1)
	g.wg.Start(func() {
		defer waitWg.Done()
		nsInf.Run(done)
	})
	return selectedGVR{
		stop: infCancel,
		wait: waitWg.Wait,
	}
}

func (g *WatchGraph) resourceEventHandlerForGVR(ctx context.Context, info gvrInfo) cache.ResourceEventHandler {
	return &resourceEventHandler{
		ctx:  ctx,
		info: info,
		setVertex: func(vid VertexID, vd VertexData, arcs arcSet[VertexID, ArcType, ArcData]) {
			g.graphMu.Lock()
			defer g.graphMu.Unlock()
			g.opts.Graph.SetVertex(vid, vd, arcs)
		},
		deleteVertex: func(vid VertexID) {
			g.graphMu.Lock()
			defer g.graphMu.Unlock()
			g.opts.Graph.DeleteVertex(vid)
		},
		gk2info: func(gk schema.GroupKind) gvrInfoLite {
			return g.gvrInfos.Load().gk2info[gk]
		},
		emitError: func(err Error) {
			select {
			case <-ctx.Done():
			case g.errCh <- err:
			}
		},
		emitWarning: func(warn Warning) {
			select {
			case <-ctx.Done():
			case g.warnCh <- warn:
			}
		},
	}
}

func (g *WatchGraph) handleNamespaceAdd(ctx context.Context, ns string) {
	g.opts.Log.Debug("Watch namespace", logz.K8sObjectName(ns))
	g.selectedNS[ns] = struct{}{}

	for _, info := range g.gvrInfos.Load().infos {
		if !info.namespaced {
			continue
		}
		g.selectedGVRs[info.gvr][ns] = g.startInformerForNamespacedGVR(ctx, info, ns)
	}
}

func (g *WatchGraph) handleNamespaceDel(ns string) {
	g.opts.Log.Debug("Unwatch namespace", logz.K8sObjectName(ns))
	delete(g.selectedNS, ns)

	waits := make([]func(), 0, len(g.selectedGVRs))
	for _, ns2data := range g.selectedGVRs {
		data := ns2data[ns]
		delete(ns2data, ns)
		data.stop()
		waits = append(waits, data.wait)
	}
	for _, w := range waits {
		w()
	}
}

func newInformer(client dynamic.Interface, gvr schema.GroupVersionResource, namespace, labelSelector, fieldSelector string) cache.SharedIndexInformer {
	nsClient := client.Resource(gvr).Namespace(namespace)
	return cache.NewSharedIndexInformerWithOptions(
		&cache.ListWatch{
			ListFunc: func(opts metav1.ListOptions) (runtime.Object, error) {
				opts.LabelSelector = labelSelector
				opts.FieldSelector = fieldSelector
				return nsClient.List(context.Background(), opts) // Do not use ctx here
			},
			WatchFunc: func(opts metav1.ListOptions) (watch.Interface, error) {
				opts.LabelSelector = labelSelector
				opts.FieldSelector = fieldSelector
				return nsClient.Watch(context.Background(), opts) // Do not use ctx - watch will be closed via Close()
			},
		},
		&unstructured.Unstructured{},
		cache.SharedIndexInformerOptions{
			ObjectDescription: gvr.String(),
		},
	)
}

func objectToEvalVars(obj *unstructured.Unstructured, gvr schema.GroupVersionResource) map[string]any {
	return map[string]any{
		"obj":         obj.Object,
		"group":       gvr.Group,
		"version":     gvr.Version,
		"resource":    gvr.Resource,
		"namespace":   obj.GetNamespace(),
		"name":        obj.GetName(),
		"labels":      obj.GetLabels(),
		"annotations": obj.GetAnnotations(),
	}
}
